@@ -2,10 +2,13 @@ import os
 import re
 import sys
 from typing import Any
+from typing import Optional
 
 import mysql.connector
 from database_infrastructure_local.number_generator import NumberGenerator
 from logger_local.MetaLogger import MetaLogger
+from python_sdk_remote.utilities import get_environment_name
+from url_remote.environment_name_enum import EnvironmentName
 from user_context_remote.user_context import UserContext
 
 from .connector import Connector
@@ -18,13 +21,17 @@ from .utils import (process_insert_data_json, process_update_data_json,
 
 class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
     """A class that provides generic CRUD functionality.
-    There are 4 main functions to create, read, update and delete data from the database.
+    There are 4 main functions to create, read, update, and delete data from the database.
     The rest of the functions are helper functions or wrappers around the main functions."""
 
+    # TODO add default_select_clause_value and default_where in all functions not only in select_multi_tuple_by_where()
     def __init__(self, *, default_schema_name: str,
                  default_table_name: str = None,
                  default_view_table_name: str = None,
                  default_id_column_name: str = None,
+                 default_select_clause_value: str = "*",
+                 default_where: str = None,
+                 # TODO: get is_test_data from the caller dirname
                  is_test_data: bool = False) -> None:
         """Initializes the GenericCRUD class. If a connection is not provided, a new connection will be created."""
         self.schema_name = default_schema_name
@@ -33,40 +40,39 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         self.default_column = default_id_column_name
         self.default_table_name = default_table_name
         self.default_view_table_name = default_view_table_name or self._get_view_name(default_table_name)
-        self.is_test_data = is_test_data
+        self.default_select_clause_value = default_select_clause_value
+        self.default_where = default_where
+        self.columns_cache = {}
+        self.is_test_data = None
+        if self.__is_test_data() or is_test_data:
+            self.is_test_data = True
         self.user_context = UserContext()
 
+    # TODO: fix ignore_duplicate
     def insert(self, *, schema_name: str = None, table_name: str = None, data_json: dict = None,
-               ignore_duplicate: bool = False) -> int:
+               ignore_duplicate: bool = False, commit_changes: bool = True) -> int:
         """Inserts a new row into the table and returns the id of the new row or -1 if an error occurred."""
 
         if ignore_duplicate:
             self.logger.warning("GenericCRUD.insert using ignore_duplicate, is it really needed?")
-        # TODO: later we may want to add a check for the table_definition to see if the table is a test table
-        # Then We will be able to remove the parameter is_test_data
-        # and only call self.__is_test_data() without the following if/else
-        if self.is_test_data:
-            if self.__is_test_data():
-                data_json["is_test_data"] = 1
-            else:
-                data_json["is_test_data"] = 0
+
         table_name = table_name or self.default_table_name
         schema_name = schema_name or self.schema_name
         self._validate_args(args=locals())
 
-        if table_name in table_definition:
-            if table_definition[table_name]["is_number_column"]:
-                view_name = self._get_view_name(table_name)
-                number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
-                data_json["number"] = number
-        else:
-            self.logger.warning(f"database-mysql-local-python generic_crud.py Table {table_name} not found in "
-                                f"database-mysql-local.table_definition_table data structure, we might need to run sql2code")
+        # if table_name in table_definition:
+        #     if table_definition[table_name]["is_number_column"]:
+        #         view_name = self._get_view_name(table_name)
+        #         number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
+        #         data_json["number"] = number
+        # else:
+        #     self.logger.warning(f"database-mysql-local-python generic_crud.py Table {table_name} not found in "
+        #                         f"database-mysql-local.table_definition_table data structure, we might need to run sql2code")
 
         # TODO: In the future we may want to check this with table_definition
-        # and not with self.is_column_in_table for better performance
-        if self.is_column_in_table(column_name="created_effective_user_id", schema_name=schema_name, table_name=table_name):
-            data_json = self.__add_create_updated_user_profile_ids(data_json=data_json, add_created_user_id=True)
+        #   and not with self.is_column_in_table for better performance
+        data_json = self.__add_create_updated_user_profile_ids(data_json=data_json, add_created_user_id=True,
+                                                               schema_name=schema_name, table_name=table_name)
 
         columns, values, data_json = process_insert_data_json(data_json=data_json)
         # We removed the IGNORE from the SQL Statement as we want to return the id of the existing row
@@ -75,11 +81,13 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                        f"VALUES ({values});"
         try:
             self.cursor.execute(insert_query, tuple(data_json.values()))
-            self.connection.commit()
+            if commit_changes:
+                self.connection.commit()
             inserted_id = self.cursor.lastrowid()
         except mysql.connector.errors.IntegrityError as exception:
             if ignore_duplicate:
-                self.logger.warning("GenericCRUD.insert: existing record found, selecting it's id")
+                self.logger.warning("GenericCRUD.insert: existing record found, selecting it's id."
+                                    f"(table_name={table_name}, data_json={data_json})")
                 inserted_id = self._get_existing_duplicate_id(table_name, exception)
             else:
                 raise exception
@@ -87,7 +95,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         return inserted_id
 
     def upsert(self, *, schema_name: str = None, table_name: str = None, view_table_name: str = None,
-               data_json: dict = None, data_json_compare: dict = None, order_by: str = None) -> int:
+               data_json: dict = None, data_json_compare: dict = None, order_by: str = None) -> Optional[int]:
         """Inserts a new row into the table, or updates an existing row if a row with the
           same values as data_json_compare exists,
           and returns the id of the new row or None if an error occurred."""
@@ -110,7 +118,8 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             params=tuple(processed_data_json_compare.values()),
             order_by=order_by)
         if table_id:
-            GenericCRUD.update_by_id(self, schema_name=schema_name, table_name=table_name, id_column_name=id_column_name,
+            GenericCRUD.update_by_id(self, schema_name=schema_name, table_name=table_name,
+                                     id_column_name=id_column_name,
                                      id_column_value=table_id, data_json=data_json)
             return table_id
         else:
@@ -139,7 +148,8 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
     def update_by_id(self, *, schema_name: str = None, table_name: str = None, id_column_name: str = None,
                      id_column_value: Any = None, data_json: dict = None,
-                     limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> None:
+                     limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None,
+                     commit_changes: bool = True) -> None:
         """Updates data in the table by ID."""
 
         table_name = table_name or self.default_table_name
@@ -154,13 +164,14 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                 extra_sql_params = (id_column_value,)
             self.update_by_where(schema_name=schema_name, table_name=table_name, where=where,
                                  data_json=data_json, params=extra_sql_params,
-                                 limit=limit, order_by=order_by)
+                                 limit=limit, order_by=order_by, commit_changes=commit_changes)
         else:
             raise Exception("Update by id requires an id_column_name")
 
     def update_by_where(self, *, schema_name: str = None, table_name: str = None, where: str = None,
                         params: tuple = None, data_json: dict = None,
-                        limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> None:
+                        limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None,
+                        commit_changes: bool = True) -> None:
         """Updates data in the table by WHERE.
         Example:
         "UPDATE table_name SET A=A_val, B=B_val WHERE C=C_val AND D=D_val"
@@ -174,8 +185,8 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         schema_name = schema_name or self.schema_name
         self._validate_args(args=locals())
 
-        if self.is_column_in_table(column_name="updated_effective_user_id", schema_name=schema_name, table_name=table_name):
-            data_json = self.__add_create_updated_user_profile_ids(data_json=data_json, add_created_user_id=False)
+        data_json = self.__add_create_updated_user_profile_ids(data_json=data_json, add_created_user_id=False,
+                                                               schema_name=schema_name, table_name=table_name)
 
         set_values, data_json = process_update_data_json(data_json)
         if not where:
@@ -189,7 +200,8 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         params = params or tuple()
         self.cursor.execute(update_query, tuple(
             data_json.values()) + params)
-        self.connection.commit()
+        if commit_changes:
+            self.connection.commit()
 
     def delete_by_id(self, *, schema_name: str = None, table_name: str = None, id_column_name: str = None,
                      id_column_value: Any = None) -> None:
@@ -225,7 +237,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
     # TODO: test distinct
     def select_one_tuple_by_id(self, *, schema_name: str = None, view_table_name: str = None,
-                               select_clause_value: str = "*",
+                               select_clause_value: str = None,
                                id_column_name: str = None,
                                id_column_value: Any = None,
                                distinct: bool = False, order_by: str = "") -> tuple:
@@ -244,7 +256,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             return tuple()
 
     def select_one_dict_by_id(self, *, schema_name: str = None, view_table_name: str = None,
-                              select_clause_value: str = "*",
+                              select_clause_value: str = None,
                               id_column_name: str = None,
                               id_column_value: Any = None,
                               distinct: bool = False, order_by: str = "") -> dict:
@@ -279,7 +291,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             return None
 
     def select_one_tuple_by_where(self, *, schema_name: str = None, view_table_name: str = None,
-                                  select_clause_value: str = "*",
+                                  select_clause_value: str = None,
                                   where: str = None, params: tuple = None,
                                   distinct: bool = False, order_by: str = "") -> tuple:
         """Selects one row from the table based on a WHERE clause and returns it as a tuple."""
@@ -297,7 +309,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             return tuple()
 
     def select_one_dict_by_where(self, *, schema_name: str = None, view_table_name: str = None,
-                                 select_clause_value: str = "*",
+                                 select_clause_value: str = None,
                                  where: str = None, params: tuple = None,
                                  distinct: bool = False, order_by: str = "") -> dict:
         """Selects one row from the table based on a WHERE clause and returns it as a dictionary."""
@@ -330,7 +342,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
             return None
 
     def select_multi_tuple_by_id(self, *, schema_name: str = None, view_table_name: str = None,
-                                 select_clause_value: str = "*",
+                                 select_clause_value: str = None,
                                  id_column_name: str = None,
                                  id_column_value: Any = None,
                                  distinct: bool = False,
@@ -360,7 +372,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                                                 order_by=order_by)
 
     def select_multi_dict_by_id(
-            self, *, schema_name: str = None, view_table_name: str = None, select_clause_value: str = "*",
+            self, *, schema_name: str = None, view_table_name: str = None, select_clause_value: str = None,
             id_column_name: str = None, id_column_value: Any = None, distinct: bool = False,
             limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> list:
         """Selects multiple rows from the table by ID and returns them as a list of dictionaries."""
@@ -376,7 +388,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
     # Old name: select_multi_by_where
     def select_multi_tuple_by_where(self, *, schema_name: str = None, view_table_name: str = None,
-                                    select_clause_value: str = "*",
+                                    select_clause_value: str = None,
                                     where: str = None, params: tuple = None, distinct: bool = False,
                                     limit: int = DEFAULT_SQL_SELECT_LIMIT,
                                     order_by: str = "") -> list:
@@ -384,6 +396,9 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
         schema_name = schema_name or self.schema_name
         view_table_name = view_table_name or self.default_view_table_name
+        select_clause_value = select_clause_value or self.default_select_clause_value
+        where = where or self.default_where
+        where = self.__where_security(where=where, view_name=view_table_name)
         self._validate_args(args=locals())
 
         # TODO: add ` to column names if they are not reserved words (like COUNT, ST_X(point), etc.)
@@ -400,7 +415,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         return result
 
     def select_multi_dict_by_where(
-            self, *, schema_name: str = None, view_table_name: str, select_clause_value: str = "*",
+            self, *, schema_name: str = None, view_table_name: str, select_clause_value: str = None,
             where: str = None, params: tuple = None, distinct: bool = False,
             limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> list:
         """Selects multiple rows from the table based on a WHERE clause and returns them as a list of dictionaries."""
@@ -415,27 +430,35 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
         return [self.convert_to_dict(row, select_clause_value) for row in result]
 
     def is_column_in_table(self, column_name: str, schema_name: str = None, table_name: str = None) -> bool:
-        """Returns True if the column is in the table, otherwise False."""
         schema_name = schema_name or self.schema_name
         table_name = table_name or self.default_table_name
         if not column_name:
             raise Exception("is_column_in_table requires a column_name")
+        columns = self.get_columns_dict(schema_name=schema_name, table_name=table_name)
+        return column_name in columns
+
+    # TODO: use table_definition_table to improve performance
+    # TODO: replace with redis in the future with table_definition_table
+    def get_columns_dict(self, schema_name: str = None, table_name: str = None) -> dict:
+        schema_name = schema_name or self.schema_name
+        table_name = table_name or self.default_table_name
+        if (schema_name, table_name) in self.columns_cache:
+            return self.columns_cache[(schema_name, table_name)]
         select_query = "SELECT column_name " \
-                       "FROM information_schema.columns " \
-                       "WHERE TABLE_SCHEMA = %s " \
-                       "AND TABLE_NAME = %s " \
-                       "AND COLUMN_NAME = %s;"
-        params = (schema_name, table_name, column_name)
+                    "FROM information_schema.columns " \
+                    "WHERE TABLE_SCHEMA = %s " \
+                    "AND TABLE_NAME = %s;"
+        params = (schema_name, table_name)
         self.cursor.execute(select_query, params)
         result = self.cursor.fetchall()
-        if result:
-            return True
-        else:
-            return False
+        columns_dict = {row[0]: None for row in result}
+        self.columns_cache[(schema_name, table_name)] = columns_dict
+        return columns_dict
 
     # helper functions:
-    def convert_to_dict(self, row: tuple, select_clause_value: str = "*") -> dict:
+    def convert_to_dict(self, row: tuple, select_clause_value: str = None) -> dict:
         """Returns a dictionary of the column names and their values."""
+        select_clause_value = select_clause_value or self.default_select_clause_value
         if select_clause_value == "*":
             column_names = [col[0] for col in self.cursor.description()]
         else:
@@ -459,7 +482,7 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
             # data_json values are allowed to contain ';', as we use them with %s (TODO: unless it's ToSQLInterface)
             if (arg_name == "data_json" and any(";" in str(x) for x in arg_value.keys()) or  # check columns
-                    arg_name != "data_json" and ";" in str(arg_value)):
+                    arg_name != "data_json" and arg_name != "params" and ";" in str(arg_value)):
                 message = f"Invalid value for {arg_name}: {arg_value} (contains ';')"
 
             if message:
@@ -472,23 +495,123 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
 
         if file_name.startswith('test_') or file_name.endswith('_test.py'):
             return True
+        elif "pytest" in file_name:
+            return True
         else:
             return False
 
-    def __add_create_updated_user_profile_ids(self, data_json: dict, add_created_user_id: bool = False) -> dict:
+    # TODO: add warning logs
+    def __add_create_updated_user_profile_ids(self, data_json: dict, add_created_user_id: bool = False,
+                                              schema_name: str = None, table_name: str = None) -> dict:
         """Adds created_user_id and updated_user_id to data_json."""
-        if add_created_user_id:
-            data_json["created_user_id"] = self.user_context.get_effective_user_id()
-            data_json["created_real_user_id"] = self.user_context.get_real_user_id()
-            data_json["created_effective_user_id"] = self.user_context.get_effective_user_id()
-            data_json["created_effective_profile_id"] = self.user_context.get_effective_profile_id()
-        data_json["updated_user_id"] = self.user_context.get_effective_user_id()
-        data_json["updated_real_user_id"] = self.user_context.get_real_user_id()
-        data_json["updated_effective_user_id"] = self.user_context.get_effective_user_id()
-        data_json["updated_effective_profile_id"] = self.user_context.get_effective_profile_id()
+        if get_environment_name() not in (EnvironmentName.DVLP1.value, EnvironmentName.PROD1.value):
+            schema_name = schema_name or self.schema_name
+            table_name = table_name or self.default_table_name
+            columns_dict = self.get_columns_dict(schema_name=schema_name, table_name=table_name)
+            if add_created_user_id:
+                if "created_user_id" in columns_dict:
+                    data_json["created_user_id"] = self.user_context.get_effective_user_id()
+                else:
+                    self.__log_warning("created_user_id", schema_name, table_name)
+                if "created_real_user_id" in columns_dict:
+                    data_json["created_real_user_id"] = self.user_context.get_real_user_id()
+                else:
+                    self.__log_warning("created_real_user_id", schema_name, table_name)
+                if "created_effective_user_id" in columns_dict:
+                    data_json["created_effective_user_id"] = self.user_context.get_effective_user_id()
+                else:
+                    self.__log_warning("created_effective_user_id", schema_name, table_name)
+                if "created_effective_profile_id" in columns_dict:
+                    data_json["created_effective_profile_id"] = self.user_context.get_effective_profile_id()
+                else:
+                    self.__log_warning("created_effective_profile_id", schema_name, table_name)
+                if "number" in columns_dict:
+                    # TODO: the commented line caused errors, we need to check it
+                    # view_name = self.default_view_table_name or self._get_view_name(table_name)
+                    view_name = table_name
+                    number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
+                    data_json["number"] = number
+                else:
+                    self.__log_warning("number", schema_name, table_name)
+                if "identifier" in columns_dict:
+                    # TODO: the commented line caused errors, we need to check it
+                    # view_name = self.default_view_table_name or self._get_view_name(table_name)
+                    view_name = table_name
+                    identifier = NumberGenerator.get_random_identifier(schema_name=schema_name, view_name=view_name,
+                                                                       identifier_column_name="identifier")
+                    data_json["identifier"] = identifier
+                else:
+                    self.__log_warning("identifier", schema_name, table_name)
+            if "updated_user_id" in columns_dict:
+                data_json["updated_user_id"] = self.user_context.get_effective_user_id()
+            else:
+                self.__log_warning("updated_user_id", schema_name, table_name)
+            if "updated_real_user_id" in columns_dict:
+                data_json["updated_real_user_id"] = self.user_context.get_real_user_id()
+            else:
+                self.__log_warning("updated_real_user_id", schema_name, table_name)
+            if "updated_effective_user_id" in columns_dict:
+                data_json["updated_effective_user_id"] = self.user_context.get_effective_user_id()
+            else:
+                self.__log_warning("updated_effective_user_id", schema_name, table_name)
+            if "updated_effective_profile_id" in columns_dict:
+                data_json["updated_effective_profile_id"] = self.user_context.get_effective_profile_id()
+            else:
+                self.__log_warning("updated_effective_profile_id", schema_name, table_name)
+            # TODO: later we may want to add a check for the table_definition to see if there is a column for is_test_data
+            if "is_test_data" in columns_dict:
+                data_json["is_test_data"] = self.is_test_data
+            else:
+                self.__log_warning("is_test_data", schema_name, table_name)
+        else:
+            schema_name = schema_name or self.schema_name
+            table_name = table_name or self.default_table_name
+            if add_created_user_id:
+                data_json["created_user_id"] = self.user_context.get_effective_user_id()
+                data_json["created_real_user_id"] = self.user_context.get_real_user_id()
+                data_json["created_effective_user_id"] = self.user_context.get_effective_user_id()
+                data_json["created_effective_profile_id"] = self.user_context.get_effective_profile_id()
+                # TODO: the commented line caused errors, we need to check it
+                # view_name = self._get_view_name(table_name)
+                view_name = table_name
+                number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
+                data_json["number"] = number
+                identifier = NumberGenerator.get_random_identifier(schema_name=schema_name, view_name=view_name,
+                                                                   identifier_column_name="identifier")
+                data_json["identifier"] = identifier
+            data_json["updated_user_id"] = self.user_context.get_effective_user_id()
+            data_json["updated_real_user_id"] = self.user_context.get_real_user_id()
+            data_json["updated_effective_user_id"] = self.user_context.get_effective_user_id()
+            data_json["updated_effective_profile_id"] = self.user_context.get_effective_profile_id()
+            data_json["is_test_data"] = self.is_test_data
         return data_json
 
-    def set_schema(self, schema_name: str or None):
+    def __log_warning(self, column_name: str, schema_name: str, table_name: str):
+        """Generates a warning log message and logs it."""
+        log_message = f"{column_name} not found in {schema_name}.{table_name}"
+        self.logger.warning(log_message=log_message)
+
+    def __where_security(self, where: str, view_name: str) -> str:
+        """Adds security to the where clause."""
+        '''
+        if self.is_column_in_table(column_name="visibility_id", schema_name=self.schema_name, table_name=view_name):
+            effective_profile_id = self.user_context.get_effective_profile_id()
+            where_security = f"(visibility_id > 1 OR created_effective_profile_id = {effective_profile_id})"
+            if not (where == "" or where is None):
+                where_security += f" AND ({where})"
+            return where_security
+        '''
+        if view_name in table_definition:
+            if table_definition[view_name].get("is_visibility"):
+                effective_profile_id = self.user_context.get_effective_profile_id()
+                where_security = f"(visibility_id > 1 OR created_effective_profile_id = {effective_profile_id})"
+                if not (where == "" or where is None):
+                    where_security += f" AND ({where})"
+                return where_security
+        return where
+
+
+    def set_schema(self, schema_name: Optional[str]):
         """Sets the given schema to be the default schema."""
         if not schema_name:
             return
@@ -507,12 +630,12 @@ class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
                 f"(user={self.connection.user}, host={self.connection.host}, schema={schema_name})")
 
     @staticmethod
-    def _get_view_name(table_name: str or None) -> str or None:
+    def _get_view_name(table_name: Optional[str]) -> Optional[str]:
         if table_name:
             return re.sub(r'(_table)$', '_view', table_name)
 
     @staticmethod
-    def generate_id_column_name(table_name: str | None) -> str | None:
+    def generate_id_column_name(table_name: Optional[str]) -> Optional[str]:
         return re.sub(r'(_table)$', '_id', table_name)
 
     def get_test_entity_id(self, *, entity_name: str, insert_function: callable, insert_kwargs: dict = None,
