@@ -1,5 +1,6 @@
 import re
 from typing import Any
+from typing import Optional
 
 from language_remote.lang_code import LangCode
 from logger_local.MetaLogger import MetaLogger
@@ -55,39 +56,78 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
 
     def add_value(self, data_ml_json: dict, table_id: int = None, lang_code: LangCode = None,
                   is_main: int = 0, data_json: dict = None, table_name: str = None,
-                  ml_table_name: str = None) -> tuple:
+                  ml_table_name: str = None) -> Optional[tuple]:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_json=data_ml_json)
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
 
-        # if this is the first insert of this data, is_main should be 1
-        if table_id is None:
-            is_main = 1
-        elif is_main == 1:
-            self._update_old_main_value_to_zero(table_id, table_name)
+        ml_table_id = None
+        old_cursor = self.cursor
+        try:
+            cursor = self.connection.cursor()
+            self.cursor = cursor
 
-        # id is the id value of the row in the table_name table
-        table_id = table_id or self.insert(data_json=data_json, ignore_duplicate=True,
-                                           table_name=table_name)
-        if table_id is None:  # TODO: delete
-            self.logger.error("Error inserting data_json", object={"data_json": str(data_json)})
+            # if this is the first insert of this data, is_main should be 1
+            if table_id is None:
+                is_main = 1
+            elif is_main == 1:
+                self._update_old_main_value_to_zero(table_id, table_name)
 
-            return None
+            # id is the id value of the row in the table_name table
+            table_id = table_id or self.insert(data_json=data_json, ignore_duplicate=True,
+                                               table_name=table_name, commit_changes=False)
+            if table_id is None:  # TODO: delete
+                self.logger.error("Error inserting data_json", object={"data_json": str(data_json)})
+                self.connection.rollback()
+                return None
 
-        if data_ml_json is None or data_ml_json == {}:
-            return table_id, None
+            if data_ml_json is None or data_ml_json == {}:
+                self.connection.commit()
+                return table_id, None
 
-        id_column_name = re.sub(r'(_table)$', '_id', table_name)
-        data_ml_json[id_column_name] = table_id
-        data_ml_json["lang_code"] = lang_code_str
-        data_ml_json["is_main"] = is_main
+            id_column_name = re.sub(r'(_table)$', '_id', table_name)
+            data_ml_json[id_column_name] = table_id
+            data_ml_json["lang_code"] = lang_code_str
+            data_ml_json["is_main"] = is_main
 
-        # ml_id is the id value of the row in the ml_table_name table
-        ml_table_id = self.insert(data_json=data_ml_json, table_name=ml_table_name, ignore_duplicate=True)
+            # ml_id is the id value of the row in the ml_table_name table
+            ml_table_id = self.insert(data_json=data_ml_json, table_name=ml_table_name, ignore_duplicate=True,
+                                      commit_changes=False)
 
+            self.connection.commit()
+        except Exception as e:
+            self.connection.rollback()
+            raise e
+        finally:
+            self.cursor = old_cursor
+            cursor.close()
         return table_id, ml_table_id
 
+    def add_value_if_not_exist(self, data_ml_json: dict, table_id: int = None, lang_code: LangCode = None,
+                               is_main: int = 0, data_json: dict = None, table_name: str = None,
+                               ml_table_name: str = None, schema_name: str = None,
+                               order_by: str = None) -> tuple:
+        self.schema_name = schema_name or self.schema_name
+        table_name = table_name or self.default_table_name
+        ml_table_name = ml_table_name or self.default_ml_table_name
+        id_column_name = GenericCRUD.generate_id_column_name(table_name)
+
+        # check if the row exists in the ml_table
+        name = data_ml_json.get("title") or data_ml_json.get("name")
+        if not name:
+            raise ValueError("name or title is required for upsert_value")
+        table_id = table_id or self.get_id_by_name(table_name=table_name, ml_table_name=ml_table_name, name=name,
+                                                   lang_code=lang_code,
+                                                   id_column_name=id_column_name, order_by=order_by)
+        ml_table_id = self.get_ml_id_by_name(ml_table_name=ml_table_name, name=name, lang_code=lang_code,
+                                             id_column_name=id_column_name, order_by=order_by)
+        if not table_id:
+            return self.add_value(data_ml_json=data_ml_json, table_id=table_id, lang_code=lang_code, is_main=is_main,
+                                  data_json=data_json, table_name=table_name, ml_table_name=ml_table_name)
+        else:
+            return table_id, ml_table_id
+
     def update_value_by_id(self, data_ml_json: dict, ml_table_id: int, table_id: int,
                            lang_code: LangCode = None, is_main: int = 0, data_json: dict = None,
                            table_name: str = None, ml_table_name: str = None,
@@ -95,7 +135,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
                            order_by: str = None) -> tuple:
 
         self.schema_name = schema_name or self.schema_name
-        lang_code_str = self._get_lang_code_str(lang_code)
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_json=data_ml_json)
         table_name = table_name or self.default_table_name
         ml_table_name = ml_table_name or self.default_ml_table_name
         id_column_name = GenericCRUD.generate_id_column_name(table_name)
@@ -105,19 +145,32 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         if ml_table_id is None:
             raise ValueError("ml_table_id is required for update_value")
 
-        if is_main == 1:
-            self._update_old_main_value_to_zero(table_id, table_name)
-        if data_json:
-            self.update_by_id(data_json=data_json, table_name=table_name, id_column_name=id_column_name,
-                              id_column_value=table_id, limit=limit, order_by=order_by)
+        old_cursor = self.cursor
+        try:
+            cursor = self.connection.cursor()
+            self.cursor = cursor
 
-        data_ml_json[id_column_name] = table_id
-        data_ml_json["lang_code"] = lang_code_str
-        data_ml_json["is_main"] = is_main
-        id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)
+            if is_main == 1:
+                self._update_old_main_value_to_zero(table_id, table_name)
+            if data_json:
+                self.update_by_id(data_json=data_json, table_name=table_name, id_column_name=id_column_name,
+                                  id_column_value=table_id, limit=limit, order_by=order_by, commit_changes=False)
 
-        self.update_by_id(data_json=data_ml_json, table_name=ml_table_name, id_column_name=id_column_name,
-                          id_column_value=ml_table_id, limit=limit, order_by=order_by)
+            data_ml_json[id_column_name] = table_id
+            data_ml_json["lang_code"] = lang_code_str
+            data_ml_json["is_main"] = is_main
+            id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)
+
+            self.update_by_id(data_json=data_ml_json, table_name=ml_table_name, id_column_name=id_column_name,
+                              id_column_value=ml_table_id, limit=limit, order_by=order_by, commit_changes=False)
+
+            self.connection.commit()
+        except Exception as e:
+            self.connection.rollback()
+            raise e
+        finally:
+            self.cursor = old_cursor
+            cursor.close()
         return table_id, ml_table_id
 
     def upsert_value(self, data_ml_json: dict, table_id: int = None, lang_code: LangCode = None,
@@ -133,15 +186,18 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         name = data_ml_json.get("title") or data_ml_json.get("name")
         if not name:
             raise ValueError("name or title is required for upsert_value")
-        table_id = self.get_id_by_name(name=name, lang_code=lang_code, id_column_name=id_column_name, order_by=order_by)
-        ml_table_id = self.get_ml_id_by_name(name=name, lang_code=lang_code, id_column_name=id_column_name,
-                                             order_by=order_by)
+        table_id = table_id or self.get_id_by_name(table_name=table_name, ml_table_name=ml_table_name, name=name,
+                                                   lang_code=lang_code,
+                                                   id_column_name=id_column_name, order_by=order_by)
+        ml_table_id = self.get_ml_id_by_name(ml_table_name=ml_table_name, name=name, lang_code=lang_code,
+                                             id_column_name=id_column_name, order_by=order_by)
         if not table_id:
             return self.add_value(data_ml_json=data_ml_json, table_id=table_id, lang_code=lang_code, is_main=is_main,
                                   data_json=data_json, table_name=table_name, ml_table_name=ml_table_name)
         else:
             return self.update_value_by_id(data_ml_json=data_ml_json, ml_table_id=ml_table_id, table_id=table_id,
-                                           lang_code=lang_code, is_main=is_main, data_json=data_json, table_name=table_name,
+                                           lang_code=lang_code, is_main=is_main, data_json=data_json,
+                                           table_name=table_name,
                                            ml_table_name=ml_table_name, schema_name=schema_name,
                                            limit=limit, order_by=order_by)
 
@@ -149,8 +205,8 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
                          id_column_name: str = None, select_clause_value: str = "*",
                          order_by: str = None) -> tuple:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
-        id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_table_name)
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
+        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(self.default_table_name)
         result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                                 params=(table_id, lang_code_str),
                                                 select_clause_value=select_clause_value,
@@ -162,29 +218,29 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
                         id_column_name: str = None,
                         order_by: str = None) -> dict:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
-        id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_table_name)
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
+        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(self.default_table_name)
         result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                                params=(table_id, lang_code_str),
-                                               order_by=f"{id_column_name} DESC")
+                                               order_by=order_by or f"{id_column_name} DESC")
 
         return result
 
     def get_main_values_tuple(self, table_id: int, id_column_name: str = None, select_clause_value: str = "*",
                               order_by: str = None) -> tuple:
 
-        id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_table_name)
+        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(self.default_table_name)
         result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                                 params=(table_id,),
                                                 select_clause_value=select_clause_value,
-                                                order_by=f"{id_column_name} DESC")
+                                                order_by=order_by or f"{id_column_name} DESC")
 
         return result
 
     def get_main_values_dict(self, table_id: int, id_column_name: str = None, select_clause_value: str = "*",
                              order_by: str = None) -> dict:
 
-        id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_table_name)
+        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(self.default_table_name)
         result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                                params=(table_id,),
                                                select_clause_value=select_clause_value,
@@ -192,18 +248,23 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
 
         return result
 
-    def get_id_by_name(self, name: str, lang_code: LangCode = None,
-                       id_column_name: str = None, order_by: str = None) -> int | None:
+    def get_id_by_name(self, name: str, table_name: str = None, ml_table_name: str = None, lang_code: LangCode = None,
+                       id_column_name: str = None, order_by: str = None) -> Optional[int]:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
-        id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_table_name)
-        if GenericCRUD.is_column_in_table(self, table_name=self.default_view_table_name, column_name="title"):
-            result = self.select_one_tuple_by_where(select_clause_value=id_column_name,
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
+        table_name = table_name or self.default_table_name
+        ml_table_name = ml_table_name or self.default_ml_table_name
+        ml_view_name = self._get_view_name(ml_table_name)
+        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(table_name)
+        if GenericCRUD.is_column_in_table(self, table_name=ml_table_name, column_name="title"):
+            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
+                                                    select_clause_value=id_column_name,
                                                     where="title=%s AND lang_code=%s",
                                                     params=(name, lang_code_str),
                                                     order_by=order_by)
         else:
-            result = self.select_one_tuple_by_where(select_clause_value=id_column_name,
+            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
+                                                    select_clause_value=id_column_name,
                                                     where="`name`=%s AND lang_code=%s",
                                                     params=(name, lang_code_str),
                                                     order_by=order_by)
@@ -212,18 +273,22 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
         else:
             return None
 
-    def get_ml_id_by_name(self, name: str, lang_code: LangCode = None,
-                          id_column_name: str = None, order_by: str = None) -> int:
+    def get_ml_id_by_name(self, name: str, ml_table_name: str = None, lang_code: LangCode = None,
+                          id_column_name: str = None, order_by: str = None) -> Optional[int]:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
-        ml_id_column_name = GenericCRUD.generate_id_column_name(id_column_name, self.default_ml_table_name)
-        if GenericCRUD.is_column_in_table(self, table_name=self.default_view_table_name, column_name="title"):
-            result = self.select_one_tuple_by_where(select_clause_value=ml_id_column_name,
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
+        ml_table_name = ml_table_name or self.default_ml_table_name
+        ml_view_name = self._get_view_name(ml_table_name)
+        ml_id_column_name = id_column_name or GenericCRUD.generate_id_column_name(ml_table_name)
+        if GenericCRUD.is_column_in_table(self, table_name=ml_table_name, column_name="title"):
+            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
+                                                    select_clause_value=ml_id_column_name,
                                                     where="title=%s AND lang_code=%s",
                                                     params=(name, lang_code_str),
                                                     order_by=order_by)
         else:
-            result = self.select_one_tuple_by_where(select_clause_value=ml_id_column_name,
+            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
+                                                    select_clause_value=ml_id_column_name,
                                                     where="`name`=%s AND lang_code=%s",
                                                     params=(name, lang_code_str),
                                                     order_by=order_by)
@@ -235,7 +300,7 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
     def delete_by_name(self, name: str,
                        lang_code: LangCode = None) -> None:
 
-        lang_code_str = self._get_lang_code_str(lang_code)
+        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
         if GenericCRUD.is_column_in_table(self, table_name=self.default_view_table_name, column_name="title"):
             self.delete_by_where(table_name=self.default_ml_table_name, where="title=%s AND lang_code=%s",
                                  params=(name, lang_code_str))
@@ -257,15 +322,21 @@ class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_COD
     def _update_old_main_value_to_zero(self, table_id: int, table_name: str) -> None:
 
         data_ml_json = {"is_main": 0}
-        id_column_name = GenericCRUD.generate_id_column_name(None, table_name)
+        id_column_name = GenericCRUD.generate_id_column_name(table_name)
         self.update_by_where(table_name=self.default_ml_table_name,
                              data_json=data_ml_json,
                              where=f"{id_column_name}=%s AND is_main=1",
                              params=(table_id,))
 
-    def _get_lang_code_str(self, lang_code: LangCode) -> str:
-        if lang_code is None:
-            lang_code_str = self.user_context.get_effective_profile_preferred_lang_code().value
+    def _get_lang_code_str(self, name: str = None, lang_code: LangCode = None,
+                           data_ml_json: dict = None) -> str:
+        if lang_code is None and name is not None:
+            lang_code_str = LangCode.detect_lang_code_str(name)
+        elif lang_code is None and data_ml_json is not None:
+            text = data_ml_json.get("title") or data_ml_json.get("name")
+            lang_code_str = LangCode.detect_lang_code_str(text)
+        elif lang_code is None:
+            lang_code_str = self.user_context.get_effective_profile_preferred_lang_code_string()
         else:
             lang_code_str = lang_code.value
         return lang_code_str
